<?php
if (!isset($_SESSION) || !is_array($_SESSION)) {
    session_start();
}
/**
 * @copyright   Copyright (C) 2023 Genius Plugin, Corp. All rights reserved.
 * @license     https://www.geniusplugin.com
 * @author      Genius Plugin
 *
 * GP Firewall is a state-of-the-art WAF solution with unparalleled protection against web-based threats.
 *
 * Main Firewall implementation.
 * In /coremio/hooks/gpfirewallHooks.php just change 'YOUR_TOKEN_HERE' in LINE N°5 with your license token

 * Free token link: https://cloud.geniusplugin.com/order.php?step=1&productGroup=1
 *
 * @return void
 */

 // In a production environment, warnings should be hidden for security reasons.
// Only comment the line below if you need to.
error_reporting(E_ALL & ~E_DEPRECATED & ~E_WARNING & ~E_NOTICE & ~E_STRICT);

class gpfirewall
{
    ############# NEVER CHANGE BELOW CODE ###############
    const WAF_V = '1.0.2';
    const WAF_TOKEN = ''; // LEAVE BLANK - DO NOT INSERT YOUR TOKEN HERE
    private $WAF_TOKEN;
    private $WAF_TOWER;
    private $WAF_HUB;
    private $WAF_CHECKPOINT;
    private $WAF_SIDE;
    private $WAF_POST;
    private $WAF_SESSION;
    private $WAF_CSRF;
    private $WAF_XFRAME;
    private $WAF_HEADERS;
    private $WAF_CALL;

    private $WAF_APP_NAME;
    private $WAF_AUTOUPDATE;
    private $WAF_update_url;
    private $WAF_update_performed;
    private $WAF_update_backup_created;
    private $WAF_update_tmp_created;
    private $WAF_update_run;
    private $WAF_update_log;

    function __construct($side = 1, $token = SELF::WAF_TOKEN)
    {

        $this->WAF_APP_NAME = 'ws';
        $this->WAF_update_url = 'https://hub.geniusplugin.com/pub/wisecp_version.php?v=';
        $this->WAF_update_run = false;
        $this->WAF_AUTOUPDATE = true; // Assume auto-update is enabled by default
        $this->WAF_update_performed = false;

        $this->WAF_CALL = false;
        if (empty($token)) return;
        $this->WAF_TOKEN = $token;
        $this->WAF_SIDE = $side;

        if (!file_exists('.waf.' . $this->WAF_TOKEN . '.json')) {
            $this->getTower();
        }

        $user_data = $this->waf_fetchData('.waf.' . $this->WAF_TOKEN . '.json');
        $user_data = json_decode($user_data);
        if (empty($user_data)) {
            $this->getTower();
            $user_data = $this->waf_fetchData('.waf.' . $this->WAF_TOKEN . '.json');
            $user_data = json_decode($user_data);
            if (!empty($user_data)) {
                if ($user_data[0] == '501') return;
                $this->WAF_TOWER = $user_data[0];
                $this->WAF_HUB = $user_data[1];
                $this->WAF_CHECKPOINT = $user_data[2];
                $this->WAF_CSRF = $user_data[3];
                $this->WAF_XFRAME = $user_data[4];
                $this->WAF_POST = $user_data[5];
                $this->WAF_SESSION = $user_data[6];
                $this->WAF_HEADERS = $user_data[7];
            } else {
                return;
            }
        } else {
            if ($user_data[0] == '501') return;
            $this->WAF_TOWER = $user_data[0];
            $this->WAF_HUB = $user_data[1];
            $this->WAF_CHECKPOINT = $user_data[2];
            $this->WAF_CSRF = $user_data[3];
            $this->WAF_XFRAME = $user_data[4];
            $this->WAF_POST = $user_data[5];
            $this->WAF_SESSION = $user_data[6];
            $this->WAF_HEADERS = $user_data[7];
            if (isset($_GET['getTower']) && $_GET['getTower'] === $token) {
                $this->getTower();
                $user_data = $this->waf_fetchData('.waf.' . $this->WAF_TOKEN . '.json');
                $user_data = json_decode($user_data);
                if (!empty($user_data)) {
                    if ($user_data[0] == '501') return;
                    $this->WAF_TOWER = $user_data[0];
                    $this->WAF_HUB = $user_data[1];
                    $this->WAF_CHECKPOINT = $user_data[2];
                    $this->WAF_CSRF = $user_data[3];
                    $this->WAF_XFRAME = $user_data[4];
                    $this->WAF_POST = $user_data[5];
                    $this->WAF_SESSION = $user_data[6];
                    $this->WAF_HEADERS = $user_data[7];
                }
            }
        }
        $this->header_modify();
        return;
    }

    /**
     * Function to Crypt data
     *
     * @param [void] $data
     * @param [string] $key
     * @return string
     */
    private function encryptClientData($data, $key)
    {
        $iv = openssl_random_pseudo_bytes(openssl_cipher_iv_length('AES-256-CBC'));
        $encrypted = openssl_encrypt($data, 'AES-256-CBC', $key, 0, $iv);
        $encoded = base64_encode($encrypted . '::' . $iv);
        return $encoded;
    }


    private function header_modify()
    {
        if (session_status() === PHP_SESSION_DISABLED || session_status()  === PHP_SESSION_NONE) {
            return;
        }

        if (!$this->WAF_CALL) {
            $this->WAF_CALL = true;
        } else {
            return;
        }
        if (empty($this->WAF_TOKEN)) return;
        if ($this->WAF_SIDE != 1 && $this->WAF_SIDE != -1) return;
        if (empty($this->WAF_TOWER)) return;
        if (empty($this->WAF_CHECKPOINT)) return;
        if ($this->WAF_CHECKPOINT != date('YmdH') . substr(date('i'), 0, 1)) {
            if ((date('YmdH') . substr(date('i'), 0, 1)) % 2 == 0) {
                $this->getTower();
            }
        }

        $dismissedKeys = array();
        $dismissedHeaders = array();

        if (isset($this->WAF_XFRAME) && !empty($this->WAF_XFRAME) && $this->WAF_XFRAME != -1) {
            $XFRAME = json_decode(base64_decode($this->WAF_XFRAME), true);
            if (is_array($XFRAME)) {
                foreach ($XFRAME as $key => $value) {
                    if ($key === 'X-Powered-By') {
                        header_remove('X-Powered-By');
                    } elseif ($key === 'Server') {
                        header_remove('Server');
                    } elseif ($key === 'enforceSessionTimeout') {
                        $this->enforceSessionTimeout($value);
                    } elseif ($key === 'enforceAbsoluteTimeout') {
                        $this->enforceAbsoluteTimeout($value);
                    } elseif ($key === 'bindSessionToIP') {
                        $this->bindSessionToIP();
                    } elseif ($key === 'bindSessionToUserAgent') {
                        $this->bindSessionToUserAgent();
                    } elseif ($key === 'privatepost') {
                        $dismissedKeys = explode(',', base64_decode($value));
                    } elseif ($key === 'privateheader') {
                        $dismissedHeaders = explode(',', base64_decode($value));
                    }  else {
                        header($key . ": " . $value);
                    }
                }
            }
        }

        if (isset($_POST['uniquehash']) && !empty($_POST['uniquehash']) && isset($_SESSION['uniquehash']) && $_SESSION['uniquehash'] === true) {
            $_SESSION['uniquehash'] = false;
            $this->checkHash(json_encode($_POST));
            exit;
        }
        $_SESSION['uniquehash'] = false;

        if (isset($this->WAF_CSRF) && !empty($this->WAF_CSRF) && $this->WAF_CSRF == 1 && $this->WAF_SIDE == 1) {
            if ($_SERVER['REQUEST_METHOD'] === 'GET' && count($_GET) > 1 && !empty($_GET)) {
                if (!isset($_GET['csrf_token']) || !isset($_GET['csrf_stamp']) || $_GET['csrf_token'] != hash('sha256', $this->WAF_TOKEN . $_GET['csrf_stamp'])) {
                    die('Invalid request.');
                }
            }
            if ($_SERVER['REQUEST_METHOD'] === 'POST' && count($_POST) > 1 && !empty($_POST)) {
                if (!isset($_POST['csrf_token']) || !isset($_POST['csrf_stamp']) || $_POST['csrf_token'] != hash('sha256', $this->WAF_TOKEN . $_POST['csrf_stamp'])) {
                    die('Invalid request.');
                }
            }

            $csrf_stamp =  substr(microtime(true) * 1000, 0, 13);
            $csrf_token = hash('sha256', $this->WAF_TOKEN . $csrf_stamp);

		?>
            <script>
                document.addEventListener("DOMContentLoaded", function() {
                    var forms = document.getElementsByTagName("form");

                    for (var i = 0; i < forms.length; i++) {
                        var form = forms[i];

                        var csrfInput = document.createElement("input");
                        csrfInput.setAttribute("type", "hidden");
                        csrfInput.setAttribute("name", "csrf_token");
                        csrfInput.setAttribute("value", "<?php echo $csrf_token; ?>");
                        form.appendChild(csrfInput);

                        var csrfInput = document.createElement("input");
                        csrfInput.setAttribute("type", "hidden");
                        csrfInput.setAttribute("name", "csrf_stamp");
                        csrfInput.setAttribute("value", "<?php echo $csrf_stamp; ?>");
                        form.appendChild(csrfInput);
                    }

                    // Get all anchor tags on the page
                    var anchorTags = document.getElementsByTagName('a');

                    // Loop through each anchor tag
                    for (var i = 0; i < anchorTags.length; i++) {
                        var href = anchorTags[i].getAttribute('href');
                        // Check if the href contains a query parameter
                        if (href.indexOf('?') !== -1) {
                            // Check if the csrf_token and csrf_stamp variables already exist in the URL
                            if (href.indexOf('csrf_token=') === -1 || href.indexOf('csrf_stamp=') === -1) {
                                // Add the csrf_token and csrf_stamp variables to the URL
                                var updatedHref = href + (href.indexOf('?') === -1 ? '?' : '&');

                                if (href.indexOf('csrf_token=') === -1) {
                                    updatedHref += 'csrf_token=' + '<?php echo $csrf_token; ?>' + '&';
                                }

                                if (href.indexOf('csrf_stamp=') === -1) {
                                    updatedHref += 'csrf_stamp=' + '<?php echo $csrf_stamp; ?>';
                                }

                                // Update the href attribute with the modified URL
                                anchorTags[i].setAttribute('href', updatedHref);
                            }
                        }
                    }

                });
            </script>
        <?php
        }

        $test_ip = '';
        $agent = '';

        $user_ip = $_SERVER['REMOTE_ADDR'];
        $server_ip = (isset($_SERVER['SERVER_ADDR'])) ? $_SERVER['SERVER_ADDR'] : '';

        if (isset($_SERVER['HTTP_Forwarded-For']) && !empty($_SERVER['HTTP_Forwarded-For'])) {
            $_SERVER['HTTP_Forwarded-For'] = trim(current(explode(',', $_SERVER['HTTP_Forwarded-For'])));
            if (filter_var($_SERVER['HTTP_Forwarded-For'], FILTER_VALIDATE_IP, FILTER_FLAG_NO_PRIV_RANGE | FILTER_FLAG_NO_RES_RANGE)) {
                $user_ip = $_SERVER['HTTP_Forwarded-For'];
            }
        }

        if (isset($_SERVER['HTTP_X-Client-IP']) && !empty($_SERVER['HTTP_X-Client-IP'])) {
            $_SERVER['HTTP_X-Client-IP'] = trim(current(explode(',', $_SERVER['HTTP_X-Client-IP'])));
            if (filter_var($_SERVER['HTTP_X-Client-IP'], FILTER_VALIDATE_IP, FILTER_FLAG_NO_PRIV_RANGE | FILTER_FLAG_NO_RES_RANGE)) {
                $user_ip = $_SERVER['HTTP_X-Client-IP'];
            }
        }

        if (isset($_SERVER['HTTP_X-Cluster-Client-Ip']) && !empty($_SERVER['HTTP_X-Cluster-Client-Ip'])) {
            $_SERVER['HTTP_X-Cluster-Client-Ip'] = trim(current(explode(',', $_SERVER['HTTP_X-Cluster-Client-Ip'])));
            if (filter_var($_SERVER['HTTP_X-Cluster-Client-Ip'], FILTER_VALIDATE_IP, FILTER_FLAG_NO_PRIV_RANGE | FILTER_FLAG_NO_RES_RANGE)) {
                $user_ip = $_SERVER['HTTP_X-Cluster-Client-Ip'];
            }
        }

        if (isset($_SERVER['HTTP_Client-IP']) && !empty($_SERVER['HTTP_Client-IP'])) {
            $_SERVER['HTTP_Client-IP'] = trim(current(explode(',', $_SERVER['HTTP_Client-IP'])));
            if (filter_var($_SERVER['HTTP_Client-IP'], FILTER_VALIDATE_IP, FILTER_FLAG_NO_PRIV_RANGE | FILTER_FLAG_NO_RES_RANGE)) {
                $user_ip = $_SERVER['HTTP_Client-IP'];
            }
        }

        if (isset($_SERVER['HTTP_CF_CONNECTING_IP']) && !empty($_SERVER['HTTP_CF_CONNECTING_IP'])) {
            $_SERVER['HTTP_CF_CONNECTING_IP'] = trim(current(explode(',', $_SERVER['HTTP_CF_CONNECTING_IP'])));
            if (filter_var($_SERVER['HTTP_CF_CONNECTING_IP'], FILTER_VALIDATE_IP, FILTER_FLAG_NO_PRIV_RANGE | FILTER_FLAG_NO_RES_RANGE)) {
                $user_ip = $_SERVER['HTTP_CF_CONNECTING_IP'];
            }
        }

        if (isset($_SERVER['HTTP_X_REAL_IP']) && !empty($_SERVER['HTTP_X_REAL_IP'])) {
            $_SERVER['HTTP_X_REAL_IP'] = trim(current(explode(',', $_SERVER['HTTP_X_REAL_IP'])));
            if (filter_var($_SERVER['HTTP_X_REAL_IP'], FILTER_VALIDATE_IP, FILTER_FLAG_NO_PRIV_RANGE | FILTER_FLAG_NO_RES_RANGE)) {
                $user_ip = $_SERVER['HTTP_X_REAL_IP'];
            }
        }

        if (isset($_SERVER['HTTP_X_FORWARDED_FOR']) && !empty($_SERVER['HTTP_X_FORWARDED_FOR'])) {
            $_SERVER['HTTP_X_FORWARDED_FOR'] = trim(current(explode(',', $_SERVER['HTTP_X_FORWARDED_FOR'])));
            if (filter_var($_SERVER['HTTP_X_FORWARDED_FOR'], FILTER_VALIDATE_IP, FILTER_FLAG_NO_PRIV_RANGE | FILTER_FLAG_NO_RES_RANGE)) {
                $user_ip = $_SERVER['HTTP_X_FORWARDED_FOR'];
            }
        }
        if ($user_ip == $server_ip || $user_ip == '::1' || $user_ip == '127.0.0.1' || $user_ip == $_SERVER['SERVER_ADDR']  || empty($user_ip)) {
            return;
        }

        if (isset($_GET['ip']) && !empty($_GET['ip'])) {
            $test_ip = $_GET['ip'];
        }
        $protocol = 'http';
        if (isset($_SERVER['HTTPS']) && $_SERVER['HTTPS'] == 443) $protocol = 'https';
        if (isset($_SERVER['SERVER_PORT']) && $_SERVER['SERVER_PORT'] == 443) $protocol = 'https';
        if (isset($_SERVER['HTTP_X_FORWARDED_PORT']) && $_SERVER['HTTP_X_FORWARDED_PORT'] == 443) $protocol = 'https';
        if (isset($_SERVER['HTTP_X_FORWARDED_PROTO']) && $_SERVER['HTTP_X_FORWARDED_PROTO'] == 'https') $protocol = 'https';
        $page = $protocol . '://' . $_SERVER['HTTP_HOST'] . $_SERVER['REQUEST_URI'];
        if (isset($_SERVER['HTTP_USER_AGENT'])) {
            $agent = base64_encode($_SERVER['HTTP_USER_AGENT']);
        }
        if ($this->WAF_POST == 1) {
            $filteredPost = $this->FilterPostData();
            /*
            User can specify which posts are inspected for malicious activity. 
            It is recommended to check all POST requests for comprehensive protection. However, the user has the option to specify which POST requests are excluded based on their specific security needs.

            $dismissedKeys[] = 'key1';
            */
            foreach ($dismissedKeys as $dismissedKey) {
                unset($filteredPost[$dismissedKey]);
            }
            $sendrequests =  $this->encryptClientData(base64_encode(json_encode($filteredPost)), $this->WAF_TOKEN);
        } else {
            $sendrequests = '';
        }
        if ($this->WAF_SESSION == 1) {
            setcookie('session_cookie', 'value', ['expires' => time() + 3600, 'path' => '/', 'domain' => '', 'secure' => true, 'httponly' => true, 'samesite' => 'Lax']);
            $sessionid =  base64_encode(session_id());
        } else {
            $sessionid = '';
        }
        if ($this->WAF_HEADERS == 1) {
            $filteredServer = getallheaders();
            if (isset($_SERVER['HTTP_ACCEPT_LANGUAGE'])) {
                $languages = explode(',', $_SERVER['HTTP_ACCEPT_LANGUAGE']);
                foreach ($languages as $lang) {
                    $lang = explode(';', $lang)[0];
                    if (preg_match('/^([a-zA-Z\-]+)/', $lang, $matches)) {
                        $preferred_language = $matches[1];
                        break;
                    }
                }
            }
            if (isset($preferred_language)) {
                $filteredServer['Detected-Language'] = $preferred_language;
            }else{
                $filteredServer['Detected-Language'] = 'ND';
            }
            // Calculate the size of the POST data string
            $filteredServer['Post-Size'] = $this->GetPostSize();
             /* 
            User can specify which headers are inspected for malicious activity.
            It is recommended to check all headers for comprehensive protection. However, the user has the option to specify which headers are excluded based on their specific security needs.
            $dismissedHeaders[] = 'HTTP_USER_AGENT';
            */
            foreach ($dismissedHeaders as $dismissedHeader) {
                unset($filteredServer[$dismissedHeader]);
            }
            $waf_headers = $this->encryptClientData(base64_encode(json_encode($filteredServer)), $this->WAF_TOKEN);
        } else {
            $waf_headers = '';
        }

        $data = array(
            'token' => $this->WAF_TOKEN,
            'ip' => $user_ip,
            'side' => $this->WAF_SIDE,
            'tip' => $test_ip,
            'ag' => $agent,
            'page' => base64_encode($page),
            'requests' => $sendrequests,
            'session' => $sessionid,
            'headers' => $waf_headers,
        );
        $url = $this->WAF_TOWER . '?app=' . $this->WAF_APP_NAME . '&v=' . self::WAF_V . '&token=' . $this->WAF_TOKEN;
        $responseText = $this->postData($url, $data);
        if ($this->WAF_update_run === true) {
            try {
                $this->waf_CheckForUpdates(); // Trigger the update check
                // print_r($this->WAF_update_log); // Error Logs for support
            } catch (Exception $e) {
                // No error message displayed for security reasons.
            }
        }

        if (!empty($responseText)) {
            $response = $responseText;
            if (strpos($responseText, '-WAFMAIL-') !== false) {
                $response = explode('-WAFMAIL-', $responseText)[0];
                $mailer = explode('-WAFMAIL-', $responseText)[1];
                $mailer = json_decode(base64_decode($mailer));
                mail($mailer[0], $mailer[1], $mailer[2]);
            }
            if (strpos($responseText, '-WAFREDIRECT-') !== false) {
                $response = explode('-WAFREDIRECT-', $responseText)[0];
                $this->getTower();
            }
            if ($response == 'deny') {
                echo null;
                exit;
            } elseif ($response == '501') {
                $this->wafJsonUpdate(json_encode(array('501')));
                return;
            } else {
                if (strpos($response, 'uniquehash') !== false) {
                    $_SESSION['uniquehash'] = true;
                }
                echo $response;
                exit;
            }
        }
    }


    /**
     * Filter POST data based on size and content.
     * @return array Returns an array of limited POST data size
     */
    private function  FilterPostData() {
        $size_mega = 1; // manually change it to decimal up to 8
        $max_total_size = $size_mega * 1024 * 1024; 
        $filtered_post = [];
        $total_size = 0;

        foreach ($_POST as $key => $value) {
            if (!is_scalar($value)) {
                continue;
            }
            $keySize = strlen($key);
            $value = strval($value);
            $valueSize = strlen($value);
            $pairSize = $keySize + $valueSize;

            // Check if adding the new key-value pair would exceed the maximum total size
            if (($total_size + $pairSize) > $max_total_size) {
                break;
            }
            $filtered_post[$key] = $value;
            $total_size += $pairSize; // Update total size with the size of the key-value pair
        }

        return $filtered_post;
    }


    /* Enforces session timeout by checking last activity and destroying session if expired
    * @param void
    * @return void
    */
    private function enforceSessionTimeout($timeoutDuration = 1800) {
        if ($this->isSessionStarted()) {

            if (!isset($_SESSION['LAST_ACTIVITY'])) {
                $_SESSION['LAST_ACTIVITY'] = time();
            }

            if (time() - $_SESSION['LAST_ACTIVITY'] > $timeoutDuration) {
                session_unset();
                session_destroy();

                if (isset($_COOKIE[session_name()])) {
                    setcookie(session_name(), '', time() - 3600, '/');
                }

                $protocol = isset($_SERVER['HTTPS']) && $_SERVER['HTTPS'] === 'on' ? 'https' : 'http';
                $mainDomain = "$protocol://" . $_SERVER['HTTP_HOST'];
                $currentUrl = "$protocol://{$_SERVER['HTTP_HOST']}{$_SERVER['REQUEST_URI']}";
                
                if ($currentUrl !== $mainDomain . '/' && $currentUrl !== $mainDomain) {
                    header("Location: $mainDomain");
                    exit();
                }
            }

            $_SESSION['LAST_ACTIVITY'] = time();
        }
    }


    /* Enforces an absolute timeout by destroying session after a set duration, regardless of activity
    * @param void
    * @return void
    */
    private function enforceAbsoluteTimeout($absoluteDuration = 7200) {
        if ($this->isSessionStarted()) {
            if (!isset($_SESSION['CREATED'])) {
                $_SESSION['CREATED'] = time();
            } elseif (time() - $_SESSION['CREATED'] > $absoluteDuration) {
                session_unset();
                session_destroy();

                if (isset($_COOKIE[session_name()])) {
                    setcookie(session_name(), '', time() - 3600, '/');
                }

                $protocol = isset($_SERVER['HTTPS']) && $_SERVER['HTTPS'] === 'on' ? 'https' : 'http';
                $mainDomain = "$protocol://" . $_SERVER['HTTP_HOST'];
                $currentUrl = "$protocol://{$_SERVER['HTTP_HOST']}{$_SERVER['REQUEST_URI']}";
                
                if ($currentUrl !== $mainDomain . '/' && $currentUrl !== $mainDomain) {
                    header("Location: $mainDomain");
                    exit();
                }
            }
        }
    }


    /* Binds session to the user's IP address to prevent hijacking from different IPs
    * @param void
    * @return void
    */
    private function bindSessionToIP() {
        if ($this->isSessionStarted()) {
            if (!isset($_SESSION['IP_ADDRESS'])) {
                $_SESSION['IP_ADDRESS'] = $_SERVER['REMOTE_ADDR'];
            } elseif ($_SESSION['IP_ADDRESS'] !== $_SERVER['REMOTE_ADDR']) {
                session_unset();
                session_destroy();

                if (isset($_COOKIE[session_name()])) {
                    setcookie(session_name(), '', time() - 3600, '/');
                }

                $protocol = isset($_SERVER['HTTPS']) && $_SERVER['HTTPS'] === 'on' ? 'https' : 'http';
                $mainDomain = "$protocol://" . $_SERVER['HTTP_HOST'];
                $currentUrl = "$protocol://{$_SERVER['HTTP_HOST']}{$_SERVER['REQUEST_URI']}";
                
                if ($currentUrl !== $mainDomain . '/' && $currentUrl !== $mainDomain) {
                    header("Location: $mainDomain");
                    exit();
                }
            }
        }
    }


    /* Binds session to the user's User-Agent to prevent hijacking from different devices or browsers
    * @param void
    * @return void
    */
    private function bindSessionToUserAgent() {
        if ($this->isSessionStarted()) {
            if (!isset($_SESSION['USER_AGENT'])) {
                $_SESSION['USER_AGENT'] = $_SERVER['HTTP_USER_AGENT'];
            } elseif ($_SESSION['USER_AGENT'] !== $_SERVER['HTTP_USER_AGENT']) {
                session_unset();
                session_destroy();

                if (isset($_COOKIE[session_name()])) {
                    setcookie(session_name(), '', time() - 3600, '/');
                }

                $protocol = isset($_SERVER['HTTPS']) && $_SERVER['HTTPS'] === 'on' ? 'https' : 'http';
                $mainDomain = "$protocol://" . $_SERVER['HTTP_HOST'];
                $currentUrl = "$protocol://{$_SERVER['HTTP_HOST']}{$_SERVER['REQUEST_URI']}";
                
                if ($currentUrl !== $mainDomain . '/' && $currentUrl !== $mainDomain) {
                    header("Location: $mainDomain");
                    exit();
                }
            }
        }
    }


    /* Checks if a session has already started to avoid multiple starts
    * @param void
    * @return bool
    */
    private function isSessionStarted() {
        if (version_compare(phpversion(), '5.4.0', '>=')) {
            return session_status() === PHP_SESSION_ACTIVE;
        } else {
            return isset($_SESSION) && count($_SESSION) > 0;
        }
    }


    /*
    Return the total size of all key-value pairs
    */
    private function GetPostSize() {
        $total_size = 0;

        foreach ($_POST as $key => $value) {
            if (is_scalar($value)) {
                $keySize = strlen($key);
                $value = strval($value);
                $valueSize = strlen($value);
                $pairSize = $keySize + $valueSize; 
                $total_size += $pairSize; 
            }
        }

        return $total_size; 
    }


    /*
    * Get (new) tower and/or (new) dashboard set for the license.
    * @return void
    *
    */
    private function getTower()
    {
        $url = 'https://hub.geniusplugin.com/license.php?v=' . self::WAF_V . '&app=' . $this->WAF_APP_NAME . '&action=gettower&token=' . $this->WAF_TOKEN;
        $result = $this->waf_fetchData($url);

        if (!empty($result)) {
            $server = json_decode($result);
            if (is_array($server)) {
                $this->WAF_TOWER = $server[0];
                $this->WAF_HUB = $server[1];
                $this->WAF_CHECKPOINT  = date('YmdH') . substr(date('i'), 0, 1);
                $this->WAF_CSRF = $server[3];
                $this->WAF_XFRAME = $server[4];
                $this->WAF_POST = $server[5];
                $this->WAF_SESSION = $server[6];
                $this->WAF_HEADERS = $server[7];
                if (isset($server[8]) && strpos($server[8], '-WAFMAIL-') === 0) {
                    $mailer = explode('-WAFMAIL-', $server[8])[1];
                    $mailer = json_decode(base64_decode($mailer));
                    mail($mailer[0], $mailer[1], $mailer[2]);
                }
                if (!file_exists('.waf.' . $this->WAF_TOKEN . '.json')) {
                    touch('.waf.' . $this->WAF_TOKEN . '.json');
                }
                $this->wafJsonUpdate(json_encode(array($this->WAF_TOWER, $this->WAF_HUB, $this->WAF_CHECKPOINT, $this->WAF_CSRF, $this->WAF_XFRAME, $this->WAF_POST, $this->WAF_SESSION, $this->WAF_HEADERS)));
                if (isset($server[9]) && $server[9] === '-WAFUPDATE-') {
                    $this->WAF_AUTOUPDATE = true;
                    // Trigger the update check $this->waf_CheckForUpdates();
                    $this->WAF_update_run = true;
                } else {
                    $this->WAF_AUTOUPDATE = false;
                }
            }
        }
    }

    private function checkHash($hashes)
    {
        $url = $this->WAF_TOWER . '?app=' . $this->WAF_APP_NAME . '&v=' . self::WAF_V;
        $data['hashes'] = $hashes;
        $responseText = $this->postData($url, $data);
        if (!empty($responseText)) {
            echo $responseText;
            exit;
        }
        return;
    }

    private function wafJsonUpdate($data)
    {
        if (!file_exists('.waf.' . $this->WAF_TOKEN . '.json')) {
            touch('.waf.' . $this->WAF_TOKEN . '.json');
        }
        $target_file = '.waf.' . $this->WAF_TOKEN . '.json';

        if ($file_name = fopen($target_file, "w")) {
            fwrite($file_name, $data);
            fclose($file_name);
            return 200;
        } else {
            return null;
        }
    }

    private function waf_CheckForUpdates()
    {
        $check_update_core_functions = $this->waf_checkRequiredCoreFunctions();
        if ($check_update_core_functions === null) {
            return;
        }

        if ($this->WAF_update_performed) {
            $this->WAF_update_log .= ' Update was already performed';
            return; // Skip update check if an update was already performed
        }

        // Check if an update has been performed in this session
        if (isset($_SESSION['update_performed']) && $_SESSION['update_performed'] === true) {
            $this->WAF_update_log .= ' Update was already performed in this session';
            return; // Skip update check if an update was already performed
        }

        // Fetch the version information from the server
        $versionInfo = $this->waf_fetchVersionInfo();

        // Check if version information is available
        if (!$versionInfo) {
            $this->WAF_update_log .= ' Version information is not available';
            return;
        }

        // Check if an update is required
        if ($this->waf_isUpdateRequired($versionInfo)) {
            // Perform the update process
            $this->waf_performUpdate($versionInfo);
        }
    }

    private function waf_fetchVersionInfo()
    {
        // Attempt to fetch version information using file_get_contents
        $versionInfoJson = $this->waf_fetchData($this->WAF_update_url . self::WAF_V . '&app=' . $this->WAF_APP_NAME.'&token='.$this->WAF_TOKEN);
        if ($versionInfoJson === false) {
            $this->WAF_update_log .= ' Fetch version information using file_get_contents failed';
            return false;
        }


        // Decode the fetched JSON version information
        $versionInfo = json_decode($versionInfoJson, true);

        // Check if the version information is valid
        if (!is_array($versionInfo) || !isset($versionInfo['new_version'], $versionInfo['download_uri'], $versionInfo['sha256'], $versionInfo['auto_update'], $versionInfo['min_php_version'], $versionInfo['min_script_version'])) {
            $this->WAF_update_log .= ' Version information invalid';
            return false;
        }
        if (empty($versionInfo['new_version']) || empty($versionInfo['download_uri']) || empty($versionInfo['sha256']) || empty($versionInfo['auto_update']) || empty($versionInfo['min_php_version']) || empty($versionInfo['min_script_version'])) {
            $this->WAF_update_log .= ' Version information invalid';
            return false;
        }
        return $versionInfo;
    }

    private function waf_isUpdateRequired($versionInfo)
    {

        // Check if auto-update is enabled
        if ($versionInfo['auto_update'] !== 1) {
            $this->WAF_update_log .= ' Auto-update is disabled';
            return false;
        }

        // Check if current PHP version meets the minimum required version
        if (version_compare(PHP_VERSION, $versionInfo['min_php_version'], '<')) {
            $this->WAF_update_log .= ' PHP version does not meet the minimum required version';
            return false;
        }

        // Check if current script meets the minimum required version for upgrade
        if (version_compare(self::WAF_V, $versionInfo['min_script_version'], '<')) {
            $this->WAF_update_log .= ' Script version does not meet the minimum required version';
            return false;
        }

        // Check if current script version is older than the new version
        if (version_compare($versionInfo['new_version'], self::WAF_V, '<')) {
            $this->WAF_update_log .= ' New version is not greater than current version';
            return false;
        }

        // Check if an autoupdate is set by user
        if ($this->WAF_AUTOUPDATE !== true) {
            $this->WAF_update_log .= ' Autoupdate is disabled';
            return false;
        }

        // Verify SHA256
        return  $this->waf_verifySha256($versionInfo['download_uri'], $versionInfo['sha256']);
    }

    private function waf_performUpdate($versionInfo)
    {
        // Download the updated zip file
        $updatedZip = @file_get_contents($versionInfo['download_uri']);

        if ($updatedZip === false) {
            // file_get_contents function alternative
            // Download the updated zip file using cURL
            $updatedZip = $this->waf_downloadWithCurl($versionInfo['download_uri']);
            if (!$updatedZip) {
                $this->WAF_update_log .= ' Download the updated zip file failed';
                return;
            }
        }


        // Create a backup directory and copy all files and folders for non-wisecp plugin
        $backupDir = __DIR__ . '/backup_';
        $this->waf_removeDirectory($backupDir);
        if ($this->WAF_APP_NAME != 'ws') {
            if ($this->WAF_update_backup_created !== true) {
                if (@mkdir($backupDir)) {
                    // Backup directory created successfully
                    $this->WAF_update_backup_created = true;

                    // Copy all files and folders
                    if ($this->waf_copyDirectoryContents(__DIR__, $backupDir) !== true) {
                        return;
                    }
                } elseif (@is_dir($backupDir)) {
                    // Backup directory created successfully
                    $this->WAF_update_backup_created = true;

                    // Copy all files and folders
                    if ($this->waf_copyDirectoryContents(__DIR__, $backupDir) !== true) {
                        return;
                    }
                } else {
                    // Failed to create backup directory
                    $this->WAF_update_log .= ' Failed to create backup directory';
                    return;
                }
            }
        }


        $tempDir = __DIR__ . '/temp_';
        $this->waf_removeDirectory($tempDir);
        // Create a temporary directory for extraction
        if ($this->WAF_update_tmp_created !== true) {
            if (@mkdir($tempDir)) {
                // Temporary directory created successfully
                $this->WAF_update_tmp_created = true;

                // Attempt to extract the updated files using PHP's ZipArchive
                $extracted = $this->waf_extractZipWithPHP($updatedZip, $tempDir, $versionInfo);

                // If extraction using PHP fails, try using PclZip
                if (!$extracted) {
                    $extracted = $this->waf_extractZipWithPclZip($updatedZip, $tempDir, $versionInfo);
                    if (!$extracted) {
                        $this->WAF_update_log .= ' Failed to extract the zip file';
                        return;
                    }
                }
            } elseif (@is_dir($backupDir)) {
                // Temporary directory created successfully
                $this->WAF_update_tmp_created = true;

                // Attempt to extract the updated files using PHP's ZipArchive
                $extracted = $this->waf_extractZipWithPHP($updatedZip, $tempDir, $versionInfo);

                // If extraction using PHP fails, try using PclZip
                if (!$extracted) {
                    $extracted = $this->waf_extractZipWithPclZip($updatedZip, $tempDir, $versionInfo);
                    if (!$extracted) {
                        $this->WAF_update_log .= ' Failed to extract the zip file';
                        return;
                    }
                }
            } else {
                // Failed to create temporary directory
                $this->WAF_update_log .= ' Failed to create temporary directory';
                return;
            }
        }

        // Copy the updated files to the main directory
        if ($this->waf_copyDirectoryContents($tempDir, __DIR__) !== true) {
            return;
        }
        $this->waf_removeDirectory($tempDir); // Remove the temporary directory

        // If a backup directory exists, restore files from backup for non-wisecp plugin
        if ($this->WAF_APP_NAME != 'ws') {
            if (file_exists($backupDir)) {
                if (!isset($_SESSION['update_performed'])  || $_SESSION['update_performed'] !== true) {
                    $this->waf_copyDirectoryContents($backupDir, __DIR__); // Restore from backup
                }
                $this->waf_removeDirectory($backupDir); // Remove the backup directory
            }
        }
        @unlink(__DIR__ . '/update.' . $versionInfo['new_version'] . '.zip'); // remove the zip file
        $this->WAF_update_performed = false;
        unset($_SESSION['update_performed']);

        // Require the updated script file & Refresh the class definition
        // require_once __FILE__;
        // new gpfirewall($this->WAF_SIDE, $this->WAF_TOKEN);

        // update the token
        $jsonFilePathPattern = './.waf.*.json';
        $hookFilePath = './coremio/hooks/gpfirewallHooks.php';
        $files = glob($jsonFilePathPattern);

        if (count($files) > 0) {
            $jsonFileName = basename($files[0]);
            $token = preg_replace('/\.waf\.(.*)\.json/', '$1', $jsonFileName);
            
            if ($token) {
                if (file_exists($hookFilePath)) {
                    $fileLines = file($hookFilePath);
                    
                    if ($fileLines !== false) {
                        foreach ($fileLines as &$line) {
                            if (strpos(trim($line), '$token =') === 0) {
                                $line = "\$token = '{$token}'; // Change YOUR_TOKEN_HERE with your license token. That's all.\n";
                                break; 
                            }
                        }
                        
                        $newContent = implode('', $fileLines);
                        $fileHandle = fopen($hookFilePath, 'w');
                        
                        if ($fileHandle !== false) {
                            fwrite($fileHandle, $newContent);
                            fclose($fileHandle);
                            
                            $this->WAF_update_log .= " Token replaced successfully.";
                        } else {
                            $this->WAF_update_log .= " Failed to open the file for writing. ";
                        }
                    } else {
                        $this->WAF_update_log .= " Failed to read gpfirewallHooks.php file. ";
                    }
                } else {
                    $this->WAF_update_log .= " gpfirewallHooks.php file not found. ";
                }
            } else {
                $this->WAF_update_log .= " Token not found in JSON file name. ";
            }
        } else {
            $this->WAF_update_log .= " Specific file not found. ";
        }
    }

    private function waf_copyDirectoryContents($source, $destination)
    {
        // Copy all files and folders from source to destination
        $dir = opendir($source);
        // Check if the destination directory is writable
        if (!is_writable($destination)) {
            $this->WAF_update_log .=  " Destination directory is not writable: $destination";
            return false;
        }

        $files_included = [
            '/coremio/hooks/gpfirewallHooks.php',
            'gpfirewall.php',
            'pclzip.php'
        ];
        $counter = 30; // Set initial counter value
        while (($file = readdir($dir)) !== false  && $counter > 0) {
            if ($file != '.' && $file != '..' && in_array($file,$files_included)) {
                $src = $source . '/' . $file;
                $dst = $destination . '/' . $file;

                // Skip copying the backup directory
                if (@is_dir($src) && strpos($file, 'backup_') === 0) {
                    continue; // Skip copying backup_ folders
                }
                // Skip copying hidden directories
                if (@is_dir($src) && substr($file, 0, 1) === '.') {
                    continue; // Skip copying hidden directories
                }

                if (@is_dir($src)) {
                    if (!@mkdir($dst) && !@is_dir($dst)) {
                        $this->WAF_update_log .=  " Failed to create directory: $dst";
                        return false;
                    }
                    $this->waf_copyDirectoryContents($src, $dst);
                } else {
                    if (!@copy($src, $dst)) {
                        $this->WAF_update_log .=  " Failed to copy file: $src to $dst";
                        return false;
                    }
                }
            }
            $counter--; // Decrement the counter
        }

        closedir($dir);
        if (strpos($source, 'temp_') !== false) {
            //After performing the update set the flag to true
            $this->WAF_update_performed = true;
            $_SESSION['update_performed'] = true;
        }
        return true;
    }

    private function waf_extractZipWithPHP($zipData, $destination, $versionInfo)
    {
        // Create a unique zip file name based on the version
        $zipFileName = $destination . '/update.' . $versionInfo['new_version'] . '.zip';

        // Attempt to use file_put_contents to save the zip file
        if (!file_put_contents($zipFileName, $zipData)) {
            // If file_put_contents fails, use an alternative method (downloadWithCurl)
            $downloadedZip = $this->waf_downloadWithCurl($versionInfo['download_uri']);
            if ($downloadedZip !== false) {
                // Attempt to save the downloaded zip data using fopen, fwrite, and fclose
                $zipFileHandle = fopen($zipFileName, 'w');
                if ($zipFileHandle !== false) {
                    fwrite($zipFileHandle, $downloadedZip);
                    fclose($zipFileHandle);
                }
            }
        }

        // Use PHP's ZipArchive to extract the files
        $zip = new ZipArchive();

        if ($zip->open($zipFileName) === true) {
            $zip->extractTo($destination);
            $zip->close();
            return true;
        } else {
            $this->WAF_update_log .= ' Failed to extract the zip file';
            return false;
        }
    }

    private function waf_fetchData($url)
    {
        // Use cURL or file_get_contents to fetch data from a URL
        if (function_exists('curl_init')) {
            $ch = curl_init($url);
            curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
            curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, 1);
            curl_setopt($ch, CURLOPT_TIMEOUT, 3);
            curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
            curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);
            curl_setopt($ch, CURLOPT_MAXREDIRS, 3);
            $data = curl_exec($ch);

            if (curl_errno($ch)) {
                if (function_exists('file_get_contents')) {
                    return @file_get_contents($url);
                } else {
                    $this->WAF_update_log .= ' Failed to fetch data from a URL 1';
                    return false;
                }
            }
            curl_close($ch);
            return $data;
        } elseif (function_exists('file_get_contents')) {
            return @file_get_contents($url);
        } else {
            $this->WAF_update_log .= ' Failed to fetch data from a URL 2';
            return false;
        }
    }

    private function waf_downloadWithCurl($url)
    {
        // Download data using cURL
        $ch = curl_init($url);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, 1);
        curl_setopt($ch, CURLOPT_TIMEOUT, 3);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
        curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);
        curl_setopt($ch, CURLOPT_MAXREDIRS, 3);
        $data = curl_exec($ch);

        if (curl_errno($ch)) {
            $this->WAF_update_log .= ' Failed to download data with curl';
            return false;
        }

        curl_close($ch);
        return $data;
    }

    private function waf_extractZipWithPclZip($zipData, $destination, $versionInfo)
    {
        require_once 'pclzip.php';  // Include the PclZip library
        // Create a unique zip file name based on the version
        $zipFileName = $destination . '/update.' . $versionInfo['new_version'] . '.zip';
        // Need file_put_contents function alternative
        file_put_contents($zipFileName, $zipData);

        // Use PclZip library to extract the files
        $zip = new PclZip($zipFileName);
        $extractionResult = $zip->extract(PCLZIP_OPT_PATH, $destination);

        if ($extractionResult === 0) {
            $this->WAF_update_log .= ' Failed to Use PclZip library to extract the files';
            return false;
        }

        return true;
    }

    private function waf_removeDirectory($path)
    {
        // Remove a directory and its contents
        if (@is_dir($path)) {
            $objects = @scandir($path);

            foreach ($objects as $object) {
                if ($object != '.' && $object != '..') {
                    if (@is_dir($path . '/' . $object)) {
                        $this->waf_removeDirectory($path . '/' . $object);
                    } else {
                        @unlink($path . '/' . $object);
                    }
                }
            }

            @rmdir($path);
        }
    }


    private function waf_verifySha256($fileUri, $expectedSha256)
    {

        $downloadedFile = @file_get_contents($fileUri);

        if ($downloadedFile === false) {
            // file_get_contents function alternative
            $downloadedFile = $this->waf_downloadWithCurl($fileUri);
            if ($downloadedFile === false) {
                $this->WAF_update_log .= ' Failed to download with curl the zip file';
                return false;
            }
        }

        // Calculate SHA256 hash and compare
        $calculatedSha256 = hash('sha256', $downloadedFile);

        if ($calculatedSha256 !== $expectedSha256) {
            $this->WAF_update_log .= ' Failed zip file checksum verification';
        }
        return $calculatedSha256 === $expectedSha256;
    }

    private function waf_checkRequiredCoreFunctions()
    {
        $requiredFunctions = array('rmdir', 'scandir', 'file_put_contents', 'mkdir', 'unlink', 'copy', 'is_dir');

        foreach ($requiredFunctions as $function) {
            if (!function_exists($function)) {
                return null;
            }
        }

        return true; // All required functions are available
    }

    private function postData($url, $data)
    {
        if (function_exists('curl_init')) {
            // Use cURL if available
            $ch = curl_init($url);
            curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
            curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, 1);
            curl_setopt($ch, CURLOPT_TIMEOUT, 3);
            curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
            curl_setopt($ch, CURLOPT_FOLLOWLOCATION, true);
            curl_setopt($ch, CURLOPT_MAXREDIRS, 3);
            curl_setopt($ch, CURLOPT_POST, true);
            curl_setopt($ch, CURLOPT_POSTFIELDS, $data);

            $response = curl_exec($ch);
            $error = curl_error($ch);

            curl_close($ch);

            if ($error) {
                return false;
            }

            return $response;
        } elseif (function_exists('stream_context_create')) {
            // Use file_get_contents with stream context if available
            $options = [
                'http' => [
                    'method' => 'POST',
                    'header' => 'Content-type: application/x-www-form-urlencoded',
                    'content' => http_build_query($data),
                ],
            ];

            $context = stream_context_create($options);
            $response = @file_get_contents($url, false, $context);

            if ($response === false) {
                return false;
            }

            return $response;
        } else {
            // Neither is available
            return false;
        }
    }
}
